from typing import Callable, Dict, Any, Optional, DefaultDict, Union

import numpy as np
from torchmetrics import Metric as TorchMetric

from baal.utils.metrics import Metrics


class MetricMixin:
    metrics: Dict[str, Union[Metrics, TorchMetric]]
    active_learning_metrics: DefaultDict[int, Dict[str, Any]]
    _active_dataset_size: int

    def active_step(self, dataset_size: Optional[int], metrics: Dict[str, Any]):
        """
        Log metrics at the end of the active learning step.

        Args:
            dataset_size: Current dataset size, if None, take state of `_active_dataset_size`.
            metrics: Metrics values

        Notes:
            Metrics can be overridden.
        """
        if dataset_size is None:
            dataset_size = self._active_dataset_size
        self.active_learning_metrics[dataset_size].update(metrics)

    def add_metric(self, name: str, initializer: Callable[[], Union[Metrics, TorchMetric]]):
        """
        Add a baal.utils.metric.Metrics or torchmetrics.Metric to the Model.

        Args:
            name (str): name of the metric.
            initializer (Callable): lambda to initialize a new instance of a
                                    Union[Metrics, TorchMetric].
        """
        self.metrics["test_" + name] = initializer()
        self.metrics["train_" + name] = initializer()

    def _reset_metrics(self, filter=""):
        """
        Reset all Metrics according to a filter.

        Args:
            filter (str): Only keep the metric if `filter` in the name.
        """
        for k, v in self.metrics.items():
            if filter in k:
                v.reset()

    def get_metrics(self, filter="") -> Dict[str, Any]:
        """
        Get all Metric values according to a filter.

        Args:
            filter (str): Only keep the metric if `filter` in the name.

        Returns:
            Dictionary with all values
        """

        def get_value(met: Union[Metrics, TorchMetric]):
            if isinstance(met, Metrics):
                return met.value
            val = met.compute().detach().cpu().numpy()
            if val.shape == ():
                val = val.item()
            return val

        metrics = {
            met_name: get_value(met) for met_name, met in self.metrics.items() if filter in met_name
        }
        if self._active_dataset_size != -1:
            # Add dataset size if it was ever set.
            metrics.update({"dataset_size": self._active_dataset_size})
        return metrics

    def _update_metrics(self, out, target, loss, filter=""):
        """
        Update all metrics.

        Args:
            out (Tensor): Prediction.
            target (Tensor): Ground truth.
            loss (Tensor): Loss from the criterion.
            filter (str): Only update metrics according to this filter.
        """
        for met_name, metric in self.metrics.items():
            if filter in met_name:
                if "loss" in met_name:
                    metric.update(loss)
                else:
                    metric.update(out, target)



from copy import deepcopy
from typing import Optional

import structlog
import torch
from baal.active.dataset.base import Dataset
from torch import nn
from torch.optim import Adam

from baal import ModelWrapper
from baal.modelwrapper import TrainingArgs
from baal.utils.metrics import ECE, ECE_PerCLs

log = structlog.get_logger("baal")


class DirichletCalibrator(object):
    """
    Adding a linear layer to a classifier model after the model is
    trained and train this new layer until convergence.
    Together with the linear layer, the model is now calibrated.
    Source: https://arxiv.org/abs/1910.12656
    Code inspired from: https://github.com/dirichletcal/experiments_neurips

    References:
        @article{kullbeyond,
                title={Beyond temperature scaling: Obtaining well-calibrated multi-class
                 probabilities with Dirichlet calibration Supplementary material},
                author={Kull, Meelis and Perello-Nieto,
                 Miquel and K{\"a}ngsepp, Markus and Silva Filho,
                  Telmo and Song, Hao and Flach, Peter}
                }

    Args:

        wrapper (ModelWrapper): Provides training and testing methods.
        num_classes (int): Number of classes in classification task.
        lr (float): Learning rate.
        reg_factor (float): Regularization factor for the linear layer weights.
        mu (float): Regularization factor for the linear layer biases.
            If not given, will be initialized by "l".
        training_duration (int): How long to train calibration layer.

    """

    def __init__(
        self,
        wrapper: ModelWrapper,
        num_classes: int,
        lr: float,
        reg_factor: float,
        mu: Optional[float] = None,
        training_duration: int = 5,
    ):
        self.num_classes = num_classes
        self.criterion = nn.CrossEntropyLoss()
        self.lr = lr
        self.reg_factor = reg_factor
        self.mu = mu or reg_factor
        self.dirichlet_linear = nn.Linear(self.num_classes, self.num_classes)
        self.model = nn.Sequential(wrapper.model, self.dirichlet_linear)
        self.optimizer = Adam(self.dirichlet_linear.parameters(), lr=self.lr)
        self.wrapper = ModelWrapper(
            self.model,
            TrainingArgs(
                criterion=self.criterion,
                optimizer=self.optimizer,
                regularizer=self.l2_reg,
                epoch=training_duration,
                use_cuda=wrapper.args.use_cuda,
            ),
        )

        self.wrapper.add_metric("ece", lambda: ECE())
        self.wrapper.add_metric("ece", lambda: ECE_PerCLs(num_classes))

    def l2_reg(self):
        """Using trainable layer's parameters for l2 regularization.

        Returns:
            The regularization term for the linear layer.
        """
        weight_p, bias_p = self.dirichlet_linear.parameters()
        w_l2_factor = weight_p.norm(2)
        b_l2_factor = bias_p.norm(2)
        return self.reg_factor * w_l2_factor + self.mu * b_l2_factor

    def calibrate(
        self,
        train_set: Dataset,
        test_set: Dataset,
        use_cuda: bool,
        double_fit: bool = False,
        **kwargs
    ):
        """
        Training the linear layer given a training set and a validation set.
        The training set should be different from what model is trained on.

        Args:
            train_set (Dataset): The training set.
            test_set (Dataset): The validation set.
            use_cuda (bool): If "True", will use GPU.
            double_fit (bool): If "True" would fit twice on the train set.
            kwargs (dict): Rest of parameters for baal.ModelWrapper.train_and_test_on_dataset().

        Returns:
            loss_history (list[float]): List of loss values for each epoch.
            model.state_dict (dict): Model weights.

        """

        # reinitialize the dirichlet calibration layer
        self.dirichlet_linear.weight.data.copy_(torch.eye(self.dirichlet_linear.weight.shape[0]))
        self.dirichlet_linear.bias.data.copy_(torch.zeros(*self.dirichlet_linear.bias.shape))
        if use_cuda:
            self.dirichlet_linear.cuda()

        loss_history, weights = self.wrapper.train_and_test_on_datasets(
            train_set, test_set, return_best_weights=True, patience=None, **kwargs
        )
        self.model.load_state_dict(weights)

        if double_fit:
            lr = self.lr / 10
            self.wrapper.args.optimizer = Adam(self.dirichlet_linear.parameters(), lr=lr)
            loss_history, weights = self.wrapper.train_and_test_on_datasets(
                train_set, test_set, return_best_weights=True, patience=None, **kwargs
            )
            self.model.load_state_dict(weights)

        return loss_history, self.model.state_dict()

    @property
    def calibrated_model(self):
        return self.model

    @property
    def metrics(self):
        return self.wrapper.metrics



from copy import deepcopy
from typing import Dict, Union, List

from numpy._typing import NDArray

from baal import ModelWrapper
from baal.active.dataset.base import Dataset
from baal.experiments import FrameworkAdapter


class ModelWrapperAdapter(FrameworkAdapter):
    def __init__(self, wrapper: ModelWrapper):
        self.wrapper = wrapper
        self._init_weight = deepcopy(self.wrapper.state_dict())

    def reset_weights(self):
        self.wrapper.load_state_dict(self._init_weight)

    def train(self, al_dataset: Dataset) -> Dict[str, float]:
        self.wrapper.train_on_dataset(al_dataset)
        return self.wrapper.get_metrics("train")

    def predict(self, dataset: Dataset, iterations: int) -> Union[NDArray, List[NDArray]]:
        return self.wrapper.predict_on_dataset(dataset, iterations=iterations)

    def evaluate(self, dataset: Dataset, average_predictions: int) -> Dict[str, float]:
        self.wrapper.test_on_dataset(dataset, average_predictions=average_predictions)
        return self.wrapper.get_metrics("test")



import warnings
from copy import deepcopy
from itertools import zip_longest
from typing import Union, Optional, Callable, Any, Dict, List, Sequence, Mapping

import numpy as np
import torch
import torch.utils.data as torchdata
from torch import Tensor

from baal.active.dataset.base import SplittedDataset, Dataset
from baal.utils.equality import deep_check

STOCHASTIC_POOL_WARNING = """
It seems that data augmentation is not disabled when iterating on the pool.
You can disable it by overriding attributes using `pool_specifics` 
when instantiating ActiveLearningDataset.
Example:
```
from torchvision.transforms import *
train_transform = Compose([Resize((224, 224)), RandomHorizontalFlip(),
                            RandomRotation(30), ToTensor()])
test_transform = Compose([Resize((224, 224)),ToTensor()])
dataset = CIFAR10(..., transform=train_transform)

al_dataset = ActiveLearningDataset(dataset,
                                    pool_specifics={'transform': test_transform})
```   
"""


def _identity(x):
    return x


class ActiveLearningDataset(SplittedDataset):
    """A dataset that allows for active learning.

    Args:
        dataset: The baseline dataset.
        labelled: An array that acts as a mask which is greater than 1 for every
            data point that is labelled, and 0 for every data point that is not
            labelled.
        make_unlabelled: The function that returns an
            unlabelled version of a datum so that it can still be used in the DataLoader.
        random_state: Set the random seed for label_randomly().
        pool_specifics: Attributes to set when creating the pool.
                                         Useful to remove data augmentation.
        last_active_steps: If specified, will iterate over the last active steps
                            instead of the full dataset. Useful when doing partial finetuning.
    """

    def __init__(
        self,
        dataset: Dataset,
        labelled: Optional[np.ndarray] = None,
        make_unlabelled: Callable = _identity,
        random_state=None,
        pool_specifics: Optional[dict] = None,
        last_active_steps: int = -1,
    ) -> None:
        self._dataset = dataset

        # The labelled_map keeps track of the step at which an item as been labelled.
        if labelled is not None:
            labelled_map: np.ndarray = labelled.astype(int)
        else:
            labelled_map = np.zeros(len(self._dataset), dtype=int)

        if pool_specifics is None:
            pool_specifics = {}
        self.pool_specifics: Dict[str, Any] = pool_specifics

        self.make_unlabelled = make_unlabelled
        # For example, FileDataset has a method 'label'. This is useful when we're in prod.
        self.can_label = self.check_dataset_can_label()
        super().__init__(
            labelled=labelled_map, random_state=random_state, last_active_steps=last_active_steps
        )
        self._warn_if_pool_stochastic()

    def check_dataset_can_label(self):
        """Check if a dataset can be labelled.

        Returns:
            Whether the dataset's label can be modified or not.

        Notes:
            To be labelled, a dataset needs a method `label`
            with definition: `label(self, idx, value)` where `value`
            is the label for indice `idx`.
        """
        has_label_attr = getattr(self._dataset, "label", None)
        if has_label_attr:
            if callable(has_label_attr):
                return True
            else:
                warnings.warn(
                    "Dataset has an attribute `label`, but it is not callable."
                    "The Dataset will not be labelled with new labels.",
                    UserWarning,
                )
        return False

    def __getitem__(self, index: int) -> Any:
        """Return items from the original dataset based on the labelled index."""
        index = self.get_indices_for_active_step()[index]
        return self._dataset[index]

    class ActiveIter:
        """Iterator over an ActiveLearningDataset."""

        def __init__(self, aldataset):
            self.i = 0
            self.aldataset = aldataset

        def __len__(self):
            return len(self.aldataset)

        def __next__(self):
            if self.i >= len(self):
                raise StopIteration

            n = self.aldataset[self.i]
            self.i = self.i + 1
            return n

    def __iter__(self):
        return self.ActiveIter(self)

    @property
    def pool(self) -> "ActiveLearningPool":
        """Returns a new Dataset made from unlabelled samples.

        Raises:
            ValueError if a pool specific attribute cannot be set.
        """
        current_dataset = deepcopy(self._dataset)

        for attr, new_val in self.pool_specifics.items():
            if hasattr(current_dataset, attr):
                setattr(current_dataset, attr, new_val)
            else:
                raise ValueError(f"{current_dataset} doesn't have {attr}")

        pool_dataset: torchdata.Subset = torchdata.Subset(
            current_dataset, (~self.labelled).nonzero()[0].reshape([-1]).tolist()
        )
        ald = ActiveLearningPool(pool_dataset, make_unlabelled=self.make_unlabelled)
        return ald

    def label(self, index: Union[list, int], value: Optional[Any] = None) -> None:
        """
        Label data points.
        The index should be relative to the pool, not the overall data.

        Args:
            index: one or many indices to label.
            value: The label value. If not provided, no modification
                                    to the underlying dataset is done.

        Raises:
            ValueError if the indices do not match the values or
             if no `value` is provided and `can_label` is True.
        """
        if isinstance(index, int):
            # We were provided only the index, we make a list.
            index_lst = [index]
            value_lst: List[Any] = [value]
        else:
            index_lst = index
            if value is None:
                value_lst = [value]
            else:
                value_lst = value

        if value_lst[0] is not None and len(index_lst) != len(value_lst):
            raise ValueError(
                "Expected `index` and `value` to be of same length when `value` is provided."
                f"Got index={len(index_lst)} and value={len(value_lst)}"
            )
        indexes = self._pool_to_oracle_index(index_lst)
        active_step = self.current_al_step + 1
        for idx, val in zip_longest(indexes, value_lst, fillvalue=None):
            if self.can_label and val is not None:
                self._dataset.label(idx, val)  # type: ignore
                self.labelled_map[idx] = active_step
            elif self.can_label and val is None:
                raise ValueError(
                    """The dataset is able to label data, but no label was provided.
                                 If this is a learn setting, please set the
                                  `ActiveLearningDataset.can_label` to `False`.
                                  """
                )
            else:
                # Regular learn usecase.
                self.labelled_map[idx] = active_step
                if val is not None:
                    warnings.warn(
                        "We will consider the original label of this datasample : {}, {}.".format(
                            self._dataset[idx][0], self._dataset[idx][1]
                        ),
                        UserWarning,
                    )

    def reset_labelled(self):
        """Reset the label map."""
        self.labelled_map = np.zeros(len(self._dataset), dtype=np.bool)

    def get_raw(self, idx: int) -> Any:
        """Get a datapoint from the underlying dataset."""
        return self._dataset[idx]

    def state_dict(self) -> Dict:
        """Return the state_dict, ie. the labelled map and random_state."""
        return {"labelled": self.labelled_map, "random_state": self.random_state}

    def load_state_dict(self, state_dict):
        """Load the labelled map and random_state with give state_dict."""
        self.labelled_map = state_dict["labelled"]
        self.random_state = state_dict["random_state"]

    def _warn_if_pool_stochastic(self):
        pool = self.pool
        if len(pool) > 0 and not deep_check(pool[0], pool[0]):
            warnings.warn(
                STOCHASTIC_POOL_WARNING,
                UserWarning,
            )


class ActiveLearningPool(torchdata.Dataset):
    """A dataset that represents the unlabelled pool for active learning.

    Args:
        dataset (Dataset): A Dataset object providing unlabelled sample.
        make_unlabelled (Callable): The function that returns an
            unlabelled version of a datum so that it can still be used in the DataLoader.

    """

    def __init__(self, dataset: torchdata.Subset, make_unlabelled: Callable = _identity) -> None:
        self._dataset: torchdata.Subset = dataset
        self.make_unlabelled = make_unlabelled

    def __getitem__(self, index: int) -> Any:
        # This datum is marked as unlabelled, so clear the label.
        return self.make_unlabelled(self._dataset[index])

    def __len__(self) -> int:
        """Return how many actual data / label pairs we have."""
        return len(self._dataset)



import copy
import warnings
from typing import Callable, Optional

import torch
from torch import nn


def replace_layers_in_module(module: nn.Module, mapping_fn: Callable, *args, **kwargs) -> bool:
    """
    Recursively iterate over the children of a module and replace them according to `mapping_fn`.

    Returns:
        True if a layer has been changed.
    """
    changed = False
    for name, child in module.named_children():
        new_module: Optional[nn.Module] = mapping_fn(child, *args, **kwargs)

        if new_module is not None:
            changed = True
            new_module.train(mode=child.training)
            module.add_module(name, new_module)

        # recursively apply to child
        changed |= replace_layers_in_module(child, mapping_fn, *args, **kwargs)
    return changed


class BayesianModule(torch.nn.Module):
    patching_function: Callable[..., torch.nn.Module]
    unpatch_function: Callable[..., torch.nn.Module]

    def __init__(self, module, *args, **kwargs):
        super().__init__()
        self.parent_module = self.__class__.patching_function(module, *args, **kwargs)

    def forward(self, *args, **kwargs):
        return self.parent_module(*args, **kwargs)

    def unpatch(self) -> torch.nn.Module:
        return self.__class__.unpatch_function(self.parent_module)

    # Context Manager
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.unpatch()


def _patching_wrapper(
    module: nn.Module,
    inplace: bool,
    patching_fn: Callable[..., Optional[nn.Module]],
    *args,
    **kwargs
) -> nn.Module:
    if not inplace:
        module = copy.deepcopy(module)
    changed = replace_layers_in_module(module, patching_fn, *args, **kwargs)
    if not changed:
        warnings.warn("No layer was modified by patch_module!", UserWarning)
    return module



from typing import Sequence, Mapping, Optional, TypeVar

import numpy as np
import torch
from torch import Tensor

T = TypeVar("T")


def deep_check(obj1, obj2) -> bool:
    if type(obj1) != type(obj2):
        return False
    elif isinstance(obj1, str):
        return bool(obj1 == obj2)
    elif isinstance(obj1, Sequence):
        return all(deep_check(i1, i2) for i1, i2 in zip(obj1, obj2))
    elif isinstance(obj1, Mapping):
        return all(deep_check(val1, obj2[key1]) for key1, val1 in obj1.items())
    elif isinstance(obj1, Tensor):
        return torch.equal(obj1, obj2)
    elif isinstance(obj1, np.ndarray):
        return bool((obj1 == obj2).all())
    else:
        return bool(obj1 == obj2)


def assert_not_none(val: Optional[T]) -> T:
    """
    This function makes sure that the variable is not None and has a fixed type for mypy purposes.
    Args:
        val: any value which is Optional.
    Returns:
        val [T]: The same value with a defined type.
    Raises:
        Assertion error if val is None.
    """
    if val is None:
        raise AssertionError(f"value of {val} is None, expected not None")
    return val



from typing import Iterable, Dict, List

import numpy as np

from baal import ActiveLearningDataset


class StoppingCriterion:
    def __init__(self, active_dataset: ActiveLearningDataset):
        self._active_ds = active_dataset

    def should_stop(self, metrics: Dict[str, float], uncertainty: Iterable[float]) -> bool:
        raise NotImplementedError


class LabellingBudgetStoppingCriterion(StoppingCriterion):
    """Stops when the labelling budget is exhausted."""

    def __init__(self, active_dataset: ActiveLearningDataset, labelling_budget: int):
        super().__init__(active_dataset)
        self._start_length = len(active_dataset)
        self.labelling_budget = labelling_budget

    def should_stop(self, metrics: Dict[str, float], uncertainty: Iterable[float]) -> bool:
        return (len(self._active_ds) - self._start_length) >= self.labelling_budget


class LowAverageUncertaintyStoppingCriterion(StoppingCriterion):
    """Stops when the average uncertainty is on average below a threshold."""

    def __init__(self, active_dataset: ActiveLearningDataset, avg_uncertainty_thresh: float):
        super().__init__(active_dataset)
        self.avg_uncertainty_thresh = avg_uncertainty_thresh

    def should_stop(self, metrics: Dict[str, float], uncertainty: Iterable[float]) -> bool:
        arr = np.array(uncertainty)
        return bool(np.mean(arr) < self.avg_uncertainty_thresh)


class EarlyStoppingCriterion(StoppingCriterion):
    """Early stopping on a particular metrics.

    Notes:
    We don't have any mandatory dependency with an early stopping implementation.
    So we have our own.
    """

    def __init__(
        self,
        active_dataset: ActiveLearningDataset,
        metric_name: str,
        patience: int = 10,
        epsilon: float = 1e-4,
    ):
        super().__init__(active_dataset)
        self.metric_name = metric_name
        self.patience = patience
        self.epsilon = epsilon
        self._acc: List[float] = []

    def should_stop(self, metrics: Dict[str, float], uncertainty: Iterable[float]) -> bool:
        self._acc.append(metrics[self.metric_name])
        near_threshold = np.isclose(np.array(self._acc), self._acc[-1], atol=self.epsilon)
        return len(near_threshold) >= self.patience and bool(
            near_threshold[-(self.patience + 1) :].all()
        )



from copy import deepcopy
from typing import Dict, cast, List, Union

from numpy._typing import NDArray

from baal.active.dataset.base import Dataset
from baal.experiments import FrameworkAdapter
from baal.transformers_trainer_wrapper import BaalTransformersTrainer


class TransformersAdapter(FrameworkAdapter):
    def __init__(self, wrapper: BaalTransformersTrainer):
        self.wrapper = wrapper
        self._init_weight = deepcopy(self.wrapper.model.state_dict())

    def reset_weights(self):
        self.wrapper.model.load_state_dict(self._init_weight)
        self.wrapper.lr_scheduler = None
        self.wrapper.optimizer = None

    def train(self, al_dataset: Dataset) -> Dict[str, float]:
        self.wrapper.train_dataset = al_dataset
        return self.wrapper.train().metrics  # type: ignore

    def predict(self, dataset: Dataset, iterations: int) -> Union[NDArray, List[NDArray]]:
        return self.wrapper.predict_on_dataset(dataset, iterations=iterations)

    def evaluate(self, dataset: Dataset, average_predictions: int) -> Dict[str, float]:
        return cast(Dict[str, float], self.wrapper.evaluate(dataset))



import sys
from collections import defaultdict
from collections.abc import Sequence
from copy import deepcopy
from dataclasses import dataclass
from typing import Callable, Optional, Union, List

import numpy as np
import structlog
import torch
from numpy._typing import NDArray
from torch.optim import Optimizer
from torch.utils.data import DataLoader
from torch.utils.data.dataloader import default_collate
from tqdm.autonotebook import tqdm

from baal.active.dataset.base import Dataset
from baal.metrics.mixin import MetricMixin
from baal.utils.array_utils import stack_in_memory
from baal.utils.cuda_utils import to_cuda
from baal.utils.equality import assert_not_none
from baal.utils.iterutils import map_on_tensor
from baal.utils.metrics import Loss
from baal.utils.warnings import raise_warnings_cache_replicated

log = structlog.get_logger("baal")


def _stack_preds(out):
    if isinstance(out[0], Sequence):
        out = [torch.stack(ts, dim=-1) for ts in zip(*out)]
    else:
        out = torch.stack(out, dim=-1)
    return out


@dataclass
class TrainingArgs:
    optimizer: Optional[Optimizer] = None
    batch_size: int = 32
    epoch: int = 0
    use_cuda: bool = torch.cuda.is_available()
    workers: int = 4
    collate_fn: Callable = default_collate
    regularizer: Optional[Callable] = None
    criterion: Optional[Callable] = None
    replicate_in_memory: bool = True


class ModelWrapper(MetricMixin):
    """
    Wrapper created to ease the training/testing/loading.

    Args:
        model (nn.Module): The model to optimize.
        args (TrainingArgs): Model arguments for training/predicting.
    """

    def __init__(self, model, args: TrainingArgs):
        self.model = model
        self.args = args
        self.metrics = dict()
        self.active_learning_metrics = defaultdict(dict)
        self.add_metric("loss", lambda: Loss())
        self._active_dataset_size = -1

        raise_warnings_cache_replicated(
            self.model, replicate_in_memory=self.args.replicate_in_memory
        )

    def train_on_dataset(self, dataset):
        """
        Train for `epoch` epochs on a Dataset `dataset.

        Args:
            dataset (Dataset): Pytorch Dataset to be trained on.

        Returns:
            The training history.
        """
        dataset_size = len(dataset)
        self.train()
        self.set_dataset_size(dataset_size)
        history = []
        log.info("Starting training", epoch=self.args.epoch, dataset=dataset_size)
        for _ in range(self.args.epoch):
            self._reset_metrics("train")
            for data, target, *_ in DataLoader(
                dataset,
                self.args.batch_size,
                True,
                num_workers=self.args.workers,
                collate_fn=self.args.collate_fn,
            ):
                _ = self.train_on_batch(data, target)
            history.append(self.get_metrics("train")["train_loss"])

        self.args.optimizer.zero_grad()  # Assert that the gradient is flushed.
        log.info("Training complete", train_loss=self.get_metrics("train")["train_loss"])
        self.active_step(dataset_size, self.get_metrics("train"))
        return history

    def test_on_dataset(
        self,
        dataset: Dataset,
        average_predictions: int = 1,
    ):
        """
        Test the model on a Dataset `dataset`.

        Args:
            dataset (Dataset): Dataset to evaluate on.
            average_predictions (int): The number of predictions to average to
                compute the test loss.

        Returns:
            Average loss value over the dataset.
        """
        self.eval()
        log.info("Starting evaluating", dataset=len(dataset))
        self._reset_metrics("test")

        for data, target, *_ in DataLoader(
            dataset,
            self.args.batch_size,
            False,
            num_workers=self.args.workers,
            collate_fn=self.args.collate_fn,
        ):
            _ = self.test_on_batch(data, target, average_predictions=average_predictions)

        log.info("Evaluation complete", test_loss=self.get_metrics("test")["test_loss"])
        self.active_step(None, self.get_metrics("test"))
        return self.get_metrics("test")["test_loss"]

    def train_and_test_on_datasets(
        self,
        train_dataset: Dataset,
        test_dataset: Dataset,
        return_best_weights=False,
        patience=None,
        min_epoch_for_es=0,
        skip_epochs=1,
    ):
        """
        Train and test the model on both Dataset `train_dataset`, `test_dataset`.

        Args:
            train_dataset (Dataset): Dataset to train on.
            test_dataset (Dataset): Dataset to evaluate on.
            return_best_weights (bool): If True, will keep the best weights and return them.
            patience (Optional[int]): If provided, will use early stopping to stop after
                                        `patience` epoch without improvement.
            min_epoch_for_es (int): Epoch at which the early stopping starts.
            skip_epochs (int): Number of epochs to skip for test_on_dataset

        Returns:
            History and best weights if required.
        """
        best_weight = None
        best_loss = 1e10
        best_epoch = 0
        hist = []
        for e in range(self.args.epoch):
            _ = self.train_on_dataset(
                train_dataset,
            )
            if e % skip_epochs == 0:
                te_loss = self.test_on_dataset(test_dataset)
                hist.append(self.get_metrics())
                if te_loss < best_loss:
                    best_epoch = e
                    best_loss = te_loss
                    if return_best_weights:
                        best_weight = deepcopy(self.state_dict())

                if patience is not None and (e - best_epoch) > patience and (e > min_epoch_for_es):
                    # Early stopping
                    break
            else:
                hist.append(self.get_metrics("train"))

        if return_best_weights:
            return hist, best_weight
        else:
            return hist

    def predict_on_dataset_generator(
        self,
        dataset: Dataset,
        iterations: int,
        half=False,
        verbose=True,
    ):
        """
        Use the model to predict on a dataset `iterations` time.

        Args:
            dataset (Dataset): Dataset to predict on.
            iterations (int): Number of iterations per sample.
            half (bool): If True use half precision.
            verbose (bool): If True use tqdm to display progress

        Notes:
            The "batch" is made of `batch_size` * `iterations` samples.

        Returns:
            Generators [batch_size, n_classes, ..., n_iterations].
        """
        self.eval()
        if len(dataset) == 0:
            return None

        log.info("Start Predict", dataset=len(dataset))
        loader = DataLoader(
            dataset,
            self.args.batch_size,
            False,
            num_workers=self.args.workers,
            collate_fn=self.args.collate_fn,
        )
        if verbose:
            loader = tqdm(loader, total=len(loader), file=sys.stdout)
        for idx, (data, *_) in enumerate(loader):

            pred = self.predict_on_batch(data, iterations)
            pred = map_on_tensor(lambda x: x.detach(), pred)
            if half:
                pred = map_on_tensor(lambda x: x.half(), pred)
            yield map_on_tensor(lambda x: x.cpu().numpy(), pred)

    def predict_on_dataset(
        self,
        dataset: Dataset,
        iterations: int,
        half=False,
        verbose=True,
    ) -> Union[NDArray, List[NDArray]]:
        """
        Use the model to predict on a dataset `iterations` time.

        Args:
            dataset (Dataset): Dataset to predict on.
            iterations (int): Number of iterations per sample.
            half (bool): If True use half precision.
            verbose (bool): If True use tqdm to show progress.

        Notes:
            The "batch" is made of `batch_size` * `iterations` samples.

        Returns:
            Array [n_samples, n_outputs, ..., n_iterations].
        """
        preds = list(
            self.predict_on_dataset_generator(
                dataset=dataset,
                iterations=iterations,
                half=half,
                verbose=verbose,
            )
        )

        if len(preds) > 0 and not isinstance(preds[0], Sequence):
            # Is an Array or a Tensor
            return np.vstack(preds)
        return [np.vstack(pr) for pr in zip(*preds)]

    def train_on_batch(self, data, target):
        """
        Train the current model on a batch using `optimizer`.

        Args:
            data (Tensor): The model input.
            target (Tensor): The ground truth.

        Returns:
            Tensor, the loss computed from the criterion.
        """

        if self.args.use_cuda:
            data, target = to_cuda(data), to_cuda(target)
        self.args.optimizer.zero_grad()
        output = self.model(data)
        loss = self.args.criterion(output, target)

        if self.args.regularizer:
            regularized_loss = loss + self.args.regularizer()
            regularized_loss.backward()
        else:
            loss.backward()

        self.args.optimizer.step()
        self._update_metrics(output, target, loss, filter="train")
        return loss

    def test_on_batch(
        self,
        data: torch.Tensor,
        target: torch.Tensor,
        average_predictions: int = 1,
    ):
        """
        Test the current model on a batch.

        Args:
            data (Tensor): The model input.
            target (Tensor): The ground truth.
            average_predictions (int): The number of predictions to average to
                compute the test loss.

        Returns:
            Tensor, the loss computed from the criterion.
        """
        with torch.no_grad():
            if self.args.use_cuda:
                data, target = to_cuda(data), to_cuda(target)

            preds = map_on_tensor(
                lambda p: p.mean(-1),
                self.predict_on_batch(data, iterations=average_predictions),
            )
            loss = assert_not_none(self.args.criterion)(preds, target)
            self._update_metrics(preds, target, loss, "test")
            return loss

    def predict_on_batch(self, data, iterations=1):
        """
        Get the model's prediction on a batch.

        Args:
            data (Tensor): The model input.
            iterations (int): Number of prediction to perform.

        Returns:
            Tensor, the loss computed from the criterion.
                    shape = {batch_size, nclass, n_iteration}.

        Raises:
            Raises RuntimeError if CUDA rans out of memory during data replication.
        """
        with torch.no_grad():
            if self.args.use_cuda:
                data = to_cuda(data)
            if self.args.replicate_in_memory:
                data = map_on_tensor(lambda d: stack_in_memory(d, iterations), data)
                try:
                    out = self.model(data)
                except RuntimeError as e:
                    raise RuntimeError(
                        """CUDA ran out of memory while BaaL tried to replicate data. See the exception above.
                    Use `replicate_in_memory=False` in order to reduce the memory requirements.
                    Note that there will be some speed trade-offs"""
                    ) from e
                out = map_on_tensor(lambda o: o.view([iterations, -1, *o.size()[1:]]), out)
                out = map_on_tensor(lambda o: o.permute(1, 2, *range(3, o.ndimension()), 0), out)
            else:
                out = [self.model(data) for _ in range(iterations)]
                out = _stack_preds(out)
            return out

    def get_params(self):
        """
        Return the parameters to optimize.

        Returns:
            Config for parameters.
        """
        return self.model.parameters()

    def state_dict(self):
        """Get the state dict(s)."""
        return self.model.state_dict()

    def load_state_dict(self, state_dict, strict=True):
        """Load the model with `state_dict`."""
        self.model.load_state_dict(state_dict, strict=strict)

    def train(self):
        """Set the model in `train` mode."""
        self.model.train()

    def eval(self):
        """Set the model in `eval mode`."""
        self.model.eval()

    def reset_fcs(self):
        """Reset all torch.nn.Linear layers."""

        def reset(m):
            if isinstance(m, torch.nn.Linear):
                m.reset_parameters()

        self.model.apply(reset)

    def reset_all(self):
        """Reset all *resetable* layers."""

        def reset(m):
            for m in self.model.modules():
                getattr(m, "reset_parameters", lambda: None)()

        self.model.apply(reset)

    def set_dataset_size(self, dataset_size: int):
        """
        Set state for dataset size. Useful for tracking.

        Args:
            dataset_size: Dataset state
        """
        self._active_dataset_size = dataset_size


def mc_inference(model, data, iterations, replicate_in_memory):
    if replicate_in_memory:
        input_shape = data.size()
        batch_size = input_shape[0]
        try:
            data = torch.stack([data] * iterations)
        except RuntimeError as e:
            raise RuntimeError(
                """CUDA ran out of memory while BaaL tried to replicate data. See the exception above.
            Use `replicate_in_memory=False` in order to reduce the memory requirements.
            Note that there will be some speed trade-offs"""
            ) from e
        data = data.view(batch_size * iterations, *input_shape[1:])
        try:
            out = model(data)
        except RuntimeError as e:
            raise RuntimeError(
                """CUDA ran out of memory while BaaL tried to replicate data. See the exception above.
            Use `replicate_in_memory=False` in order to reduce the memory requirements.
            Note that there will be some speed trade-offs"""
            ) from e
        out = map_on_tensor(lambda o: o.view([iterations, batch_size, *o.size()[1:]]), out)
        out = map_on_tensor(lambda o: o.permute(1, 2, *range(3, o.ndimension()), 0), out)
    else:
        out = [model(data) for _ in range(iterations)]
        if isinstance(out[0], Sequence):
            out = [torch.stack(ts, dim=-1) for ts in zip(*out)]
        else:
            out = torch.stack(out, dim=-1)
    return out



import itertools
from typing import Union, Optional, Any

import numpy as np
import structlog
from torch.utils.data import Subset
from tqdm.autonotebook import tqdm

from baal import ModelWrapper, ActiveLearningDataset
from baal.active.dataset.base import Dataset
from baal.active.heuristics import AbstractHeuristic
from baal.active.stopping_criteria import StoppingCriterion, LabellingBudgetStoppingCriterion
from baal.experiments import FrameworkAdapter
from baal.experiments.modelwrapper import ModelWrapperAdapter

try:
    import transformers
    from baal.transformers_trainer_wrapper import BaalTransformersTrainer
    from baal.experiments.transformers import TransformersAdapter

    TRANSFORMERS_AVAILABLE = True
except ImportError:
    BaalTransformersTrainer = None  # type: ignore
    TransformersAdapter = None  # type: ignore
    TRANSFORMERS_AVAILABLE = False

log = structlog.get_logger("baal")


class ActiveLearningExperiment:
    """Experiment manager for Baal.

    Takes care of:
        1. Train the model on the label set.
        2. Evaluate the model on the evaluation set.
        3. Predict on the unlabelled examples.
        4. Label the most uncertain examples.
        5. Stop the experiment if finished.

    Args:
        trainer: Huggingface or ModelWrapper to train
        al_dataset: Active learning dataset
        eval_dataset: Evaluation Dataset
        heuristic: Uncertainty heuristic to use
        query_size: How many items to label per round.
        iterations: How many MC sampling to perform per item.
        pool_size: Optionally limit the size of the unlabelled pool.
        criterion: Stopping criterion for the experiment.
    """

    def __init__(
        self,
        trainer: Union[ModelWrapper, "BaalTransformersTrainer"],
        al_dataset: ActiveLearningDataset,
        eval_dataset: Dataset,
        heuristic: AbstractHeuristic,
        query_size: int = 100,
        iterations: int = 20,
        pool_size: Optional[int] = None,
        criterion: Optional[StoppingCriterion] = None,
    ):
        self.al_dataset = al_dataset
        self.eval_dataset = eval_dataset
        self.heuristic = heuristic
        self.query_size = query_size
        self.iterations = iterations
        self.criterion = criterion or LabellingBudgetStoppingCriterion(
            al_dataset, labelling_budget=al_dataset.n_unlabelled
        )
        self.pool_size = pool_size
        self.adapter = self._get_adapter(trainer)

    def start(self):
        records = []
        _start = len(self.al_dataset)
        if _start == 0:
            raise ValueError(
                "No item labelled in the training set."
                " Did you run `ActiveLearningDataset.label_randomly`?"
            )
        for _ in tqdm(
            itertools.count(start=0),  # Infinite counter to rely on Criterion
            desc="Active Experiment",
            # Upper bound estimation.
            total=np.round(self.al_dataset.n_unlabelled // self.query_size),
        ):
            self.adapter.reset_weights()
            train_metrics = self.adapter.train(self.al_dataset)
            eval_metrics = self.adapter.evaluate(
                self.eval_dataset, average_predictions=self.iterations
            )
            pool = self._get_pool()
            ranks, uncertainty = self.heuristic.get_ranks(
                self.adapter.predict(pool, iterations=self.iterations)
            )
            self.al_dataset.label(ranks[: self.query_size])
            records.append({**train_metrics, **eval_metrics})
            if self.criterion.should_stop(eval_metrics, uncertainty):
                log.info("Experiment complete", num_labelled=len(self.al_dataset) - _start)
                return records

    def _get_adapter(
        self, trainer: Union[ModelWrapper, "BaalTransformersTrainer"]
    ) -> FrameworkAdapter:
        if isinstance(trainer, ModelWrapper):
            return ModelWrapperAdapter(trainer)
        elif TRANSFORMERS_AVAILABLE and isinstance(trainer, BaalTransformersTrainer):
            return TransformersAdapter(trainer)
        raise ValueError(
            f"{type(trainer)} is not a supported trainer."
            " Baal supports ModelWrapper and BaalTransformersTrainer"
        )

    def _get_pool(self):
        if self.pool_size is None:
            return self.al_dataset.pool
        pool = self.al_dataset.pool
        if len(pool) < self.pool_size:
            return pool
        indices = np.random.choice(len(pool), min(len(pool), self.pool_size), replace=False)
        return Subset(pool, indices)



from typing import Optional

import torch
from torch import nn, Tensor

from baal.bayesian.common import BayesianModule, _patching_wrapper


class LRUCacheModule(nn.Module):
    def __init__(self, module, size=1):
        super().__init__()
        if size != 1:
            raise ValueError("We do not support LRUCache bigger than 1.")
        self.module = module
        self._memory_input = None
        self._memory_output = None

    def _is_cache_void(self, x):
        return self._memory_input is None or not torch.equal(self._memory_input, x)

    def __call__(self, x: Tensor):
        if self.training:
            return self.module(x)
        if self._is_cache_void(x):
            self._memory_input = x
            self._memory_output = self.module(x)
        return self._memory_output


def _caching_mapping_fn(module: torch.nn.Module) -> Optional[nn.Module]:
    new_module: Optional[nn.Module] = None
    # Could add more
    if isinstance(module, (nn.Linear, nn.Conv2d)):
        new_module = LRUCacheModule(module=module)
    return new_module


def _caching_unmapping_fn(module: torch.nn.Module) -> Optional[nn.Module]:
    new_module: Optional[nn.Module] = None

    if isinstance(module, LRUCacheModule):
        new_module = module.module
    return new_module


def patch_module(module: torch.nn.Module, inplace: bool = True) -> torch.nn.Module:
    return _patching_wrapper(module, inplace=inplace, patching_fn=_caching_mapping_fn)


def unpatch_module(module: torch.nn.Module, inplace: bool = True) -> torch.nn.Module:
    return _patching_wrapper(module, inplace=inplace, patching_fn=_caching_unmapping_fn)


class MCCachingModule(BayesianModule):
    patching_function = patch_module
    unpatch_function = unpatch_module



import abc
from typing import Dict, List, Union

from numpy._typing import NDArray

from baal.active.dataset.base import Dataset


class FrameworkAdapter(abc.ABC):
    def reset_weights(self):
        raise NotImplementedError

    def train(self, al_dataset: Dataset) -> Dict[str, float]:
        raise NotImplementedError

    def predict(self, dataset: Dataset, iterations: int) -> Union[NDArray, List[NDArray]]:
        raise NotImplementedError

    def evaluate(self, dataset: Dataset, average_predictions: int) -> Dict[str, float]:
        raise NotImplementedError



"""Some heuristics are simply combinations of others."""
from typing import Union, Callable

from baal.active.heuristics import BALD, Entropy
from baal.active.heuristics.stochastics import PowerSampling


def PowerBALD(query_size: int, temperature: float = 1.0, reduction: Union[str, Callable] = "none"):
    return PowerSampling(BALD(reduction=reduction), query_size=query_size, temperature=temperature)


def PowerEntropy(
    query_size: int, temperature: float = 1.0, reduction: Union[str, Callable] = "none"
):
    return PowerSampling(
        Entropy(reduction=reduction), query_size=query_size, temperature=temperature
    )



import copy
import warnings
from typing import List, Optional, cast, Dict
from baal.bayesian.common import replace_layers_in_module, _patching_wrapper, BayesianModule
import torch
from torch import nn
from packaging.version import parse as parse_version


Sequence = List[str]


def get_weight_drop_module(name: str, weight_dropout, **kwargs):
    return {"Conv2d": WeightDropConv2d, "Linear": WeightDropLinear}[name](weight_dropout, **kwargs)


class WeightDropMixin:
    _kwargs: Dict

    def unpatch(self):
        new_module = self.__class__.__bases__[0](**self._kwargs)
        new_module.load_state_dict(self.state_dict())
        return new_module


class WeightDropLinear(torch.nn.Linear, WeightDropMixin):
    """
    Thanks to PytorchNLP for the initial implementation
    # code from https://pytorchnlp.readthedocs.io/en/latest/_modules/torchnlp/nn/weight_drop.html
    of class WeightDropLinear. Their `License
    <https://github.com/PetrochukM/PyTorch-NLP/blob/master/LICENSE>`__.
    Wrapper around :class:`torch.nn.Linear` that adds ``weight_dropout`` named argument.

    Args:
        weight_dropout (float): The probability a weight will be dropped.
    """

    def __init__(self, weight_dropout=0.0, **kwargs):
        wanted = ["in_features", "out_features"]
        self._kwargs = {k: v for k, v in kwargs.items() if k in wanted}
        super().__init__(**self._kwargs)
        self._weight_dropout = weight_dropout

    def forward(self, input):
        w = torch.nn.functional.dropout(self.weight, p=self._weight_dropout, training=True)
        return torch.nn.functional.linear(input, w, self.bias)


class WeightDropConv2d(torch.nn.Conv2d, WeightDropMixin):
    """
    Reimplemmentation of WeightDrop for Conv2D. Thanks to PytorchNLP for the initial implementation
    of class WeightDropLinear. Their `License
    <https://github.com/PetrochukM/PyTorch-NLP/blob/master/LICENSE>`__.
    Wrapper around :class: 'torch.nn.Conv' that adds '' weight_dropout '' named argument.

    Args:
        weight_dropout (float): The probability a weight will be dropped.
    """

    def __init__(self, weight_dropout=0.0, **kwargs):
        wanted = ["in_channels", "out_channels", "kernel_size", "dilation", "padding"]
        self._kwargs = {k: v for k, v in kwargs.items() if k in wanted}
        super().__init__(**self._kwargs)
        self._weight_dropout = weight_dropout
        self._torch_version = parse_version(torch.__version__)

    def forward(self, input):
        kwargs = {
            "input": input,
            "weight": torch.nn.functional.dropout(
                self.weight, p=self._weight_dropout, training=True
            ),
        }
        if self._torch_version >= parse_version("1.8.0"):
            # Bias was added as a required argument in this version.
            kwargs["bias"] = self.bias
        return self._conv_forward(**kwargs)


def patch_module(
    module: torch.nn.Module, layers: Sequence, weight_dropout: float = 0.0, inplace: bool = True
) -> torch.nn.Module:
    """
    Replace given layers with weight_drop module of that layer.

    Args:
        module : torch.nn.Module
            The module in which you would like to replace dropout layers.
        layers : list[str]
            Name of layers to be replaced from ['Conv', 'Linear', 'LSTM', 'GRU'].
        weight_dropout (float): The probability a weight will be dropped.
        inplace : bool, optional
            Whether to modify the module in place or return a copy of the module.

    Returns:
        torch.nn.Module:
            The modified module, which is either the same object as you passed in
            (if inplace = True) or a copy of that object.
    """
    return _patching_wrapper(
        module,
        inplace=inplace,
        patching_fn=_dropconnect_mapping_fn,
        layers=layers,
        weight_dropout=weight_dropout,
    )


def unpatch_module(module: torch.nn.Module, inplace: bool = True) -> torch.nn.Module:
    """
    Unpatch Dropconnect module to recover initial module.

    Args:
        module (torch.nn.Module):
            The module in which you would like to replace dropout layers.
        inplace (bool, optional):
            Whether to modify the module in place or return a copy of the module.

    Returns:
        torch.nn.Module
            The modified module, which is either the same object as you passed in
            (if inplace = True) or a copy of that object.
    """
    return _patching_wrapper(module, inplace=inplace, patching_fn=_droconnect_unmapping_fn)


def _dropconnect_mapping_fn(module: torch.nn.Module, layers, weight_dropout) -> Optional[nn.Module]:
    new_module: Optional[nn.Module] = None
    for layer in layers:
        if isinstance(module, getattr(torch.nn, layer)):
            new_module = get_weight_drop_module(layer, weight_dropout, **module.__dict__)
            break
    if isinstance(module, nn.Dropout):
        module._baal_p: float = module.p  # type: ignore
        module.p = 0.0
    return new_module


def _droconnect_unmapping_fn(module: torch.nn.Module) -> Optional[nn.Module]:
    new_module: Optional[nn.Module] = None
    if isinstance(module, WeightDropMixin):
        new_module = module.unpatch()

    if isinstance(module, nn.Dropout):
        module.p = module._baal_p  # type: ignore

    return new_module


class MCDropoutConnectModule(BayesianModule):
    """Create a module that with all dropout layers patched.
    With MCDropoutConnectModule, it could be decided which type of modules to be
    replaced.

    Args:
        module (torch.nn.Module):
            A fully specified neural network.
        layers (list[str]):
            Name of layers to be replaced from ['Conv', 'Linear', 'LSTM', 'GRU'].
        weight_dropout (float): The probability a weight will be dropped.
    """

    patching_function = patch_module
    unpatch_function = unpatch_module



import types

import numpy as np
import structlog
from scipy.special import softmax
from scipy.stats import rankdata

from baal.active.heuristics import AbstractHeuristic, Sequence

log = structlog.get_logger("baal")
EPSILON = 1e-8


class StochasticHeuristic(AbstractHeuristic):
    def __init__(self, base_heuristic: AbstractHeuristic, query_size):
        """Heuristic that is stochastic to improve diversity.

        Common acquisition functions are heavily impacted by duplicates.
        When using a `top-k` approache where the most
        uncertain examples are selected, the acquisition function can select many duplicates.
        Techniques such as BADGE (Ash et al, 2019) or BatchBALD (Kirsh et al. 2019)
        are common solutions to this problem, but they are quite expensive.

        Stochastic acquisitions are cheap to compute and get similar performances.

        References:
            Stochastic Batch Acquisition for Deep Active Learning, Kirsch et al. (2022)
            https://arxiv.org/abs/2106.12059

        Args:
            base_heuristic: Heuristic to get uncertainty from before sampling.
            query_size: These heuristics will return `query_size` items.
        """
        # TODO handle reverse
        super().__init__(reverse=False)
        self._bh = base_heuristic
        self.query_size = query_size

    def get_ranks(self, predictions):
        # Get the raw uncertainty from the base heuristic.
        scores = self.get_scores(predictions)
        # Create the distribution to sample from.
        distributions = self._make_distribution(scores)
        # Force normalization for np.random.choice
        distributions = np.clip(distributions, 0, None)
        distributions /= distributions.sum()

        # TODO Seed?
        if (distributions > 0).sum() < self.query_size:
            log.warn("Not enough values, return random")
            distributions = np.ones_like(distributions) / len(distributions)
        return (
            np.random.choice(len(distributions), self.query_size, replace=False, p=distributions),
            distributions,
        )

    def get_scores(self, predictions):
        if isinstance(predictions, types.GeneratorType):
            scores = self._bh.get_uncertainties_generator(predictions)
        else:
            scores = self._bh.get_uncertainties(predictions)
        if isinstance(scores, Sequence):
            scores = np.concatenate(scores)
        return scores

    def _make_distribution(self, scores: np.ndarray) -> np.ndarray:
        raise NotImplementedError


class PowerSampling(StochasticHeuristic):
    def __init__(self, base_heuristic: AbstractHeuristic, query_size, temperature=1.0):
        """Samples from the uncertainty distribution without modification beside
        temperature scaling and normalization.

        Stochastic heuristic that assumes that the uncertainty distribution
        is positive and that items with near-zero uncertainty are uninformative.
        Empirically worked the best in the paper.

        References:
            Stochastic Batch Acquisition for Deep Active Learning, Kirsch et al. (2022)
            https://arxiv.org/abs/2106.12059

        Args:
            base_heuristic: Heuristic to get uncertainty from before sampling.
            query_size: These heuristics will return `query_size` items.
            temperature: Value to temper the uncertainty distribution before sampling.
        """
        super().__init__(base_heuristic=base_heuristic, query_size=query_size)
        self.temperature = temperature

    def _make_distribution(self, scores: np.ndarray) -> np.ndarray:
        scores = scores ** (1 / self.temperature)
        scores = scores / scores.sum()
        return scores


class GibbsSampling(StochasticHeuristic):
    def __init__(self, base_heuristic: AbstractHeuristic, query_size, temperature=1.0):
        """Samples from the uncertainty distribution after applying softmax.

        References:
            Stochastic Batch Acquisition for Deep Active Learning, Kirsch et al. (2022)
            https://arxiv.org/abs/2106.12059

        Args:
            base_heuristic: Heuristic to get uncertainty from before sampling.
            query_size: These heuristics will return `query_size` items.
            temperature: Value to temper the uncertainty distribution before sampling.
        """
        super().__init__(base_heuristic=base_heuristic, query_size=query_size)
        self.temperature = temperature

    def _make_distribution(self, scores: np.ndarray) -> np.ndarray:
        scores /= self.temperature
        # scores dimensions is [N]
        scores = softmax(scores)
        return scores


class RankBasedSampling(StochasticHeuristic):
    def __init__(self, base_heuristic: AbstractHeuristic, query_size, temperature=1.0):
        """Samples from the ranks of the uncertainty distribution.

        References:
            Stochastic Batch Acquisition for Deep Active Learning, Kirsch et al. (2022)
            https://arxiv.org/abs/2106.12059

        Args:
            base_heuristic: Heuristic to get uncertainty from before sampling.
            query_size: These heuristics will return `query_size` items.
            temperature: Value to temper the uncertainty distribution before sampling.
        """
        super().__init__(base_heuristic=base_heuristic, query_size=query_size)
        self.temperature = temperature

    def _make_distribution(self, scores: np.ndarray) -> np.ndarray:
        rank = rankdata(-scores)
        weights = rank ** (-1 / self.temperature)
        normalized_weights: np.ndarray = weights / weights.sum()
        return normalized_weights



import warnings

from torch import nn

from baal.bayesian.caching_utils import LRUCacheModule

WARNING_CACHE_REPLICATED = """
To use MCCachingModule at maximum effiency, we recommend using
 `replicate_in_memory=False`, but it is `True`.
"""


def raise_warnings_cache_replicated(module, replicate_in_memory):
    if (
        isinstance(module, nn.Module)
        and replicate_in_memory
        and any(isinstance(m, LRUCacheModule) for m in module.modules())
    ):
        warnings.warn(WARNING_CACHE_REPLICATED, UserWarning)



from itertools import zip_longest
from typing import Tuple, Optional, Any, Union

import numpy as np

from baal.active.dataset.base import SplittedDataset


class ActiveNumpyArray(SplittedDataset):
    """
    Active dataset for numpy arrays. Useful when using sklearn.

    Args:
        dataset (Tuple[ndarray, ndarray]): [Train x, train y], The dataset.
        labelled (Union[np.ndarray, torch.Tensor]):
            An array/tensor that acts as a boolean mask which is True for every
            data point that is labelled, and False for every data point that is not
            labelled.
        random_state: Random seed for the ActiveLearningDataset
    """

    def __init__(
        self,
        dataset: Tuple[np.ndarray, np.ndarray],
        labelled: Optional[np.ndarray] = None,
        random_state: Any = None,
    ) -> None:
        self._dataset = dataset
        # The labelled_map keeps track of the step at which an item as been labelled.
        if labelled is not None:
            labelled_map: np.ndarray = labelled.astype(int)
        else:
            labelled_map = np.zeros(len(self._dataset[0]), dtype=int)
        super().__init__(labelled=labelled_map, random_state=random_state, last_active_steps=-1)

    @property
    def pool(self):
        """Return the unlabelled portion of the dataset."""
        return self._dataset[0][~self.labelled], self._dataset[1][~self.labelled]

    @property
    def dataset(self):
        """Return the labelled portion of the dataset."""
        return self._dataset[0][self.labelled], self._dataset[1][self.labelled]

    def get_raw(self, idx: int) -> Any:
        return self._dataset[0][idx], self._dataset[1][idx]

    def __iter__(self):
        return zip(*self._dataset)

    def __getitem__(self, index):
        index = self.get_indices_for_active_step()[index]
        return self._dataset[0][index], self._dataset[1][index]

    def label(self, index: Union[list, int], value: Optional[Any] = None) -> None:
        """
        Label data points.
        The index should be relative to the pool, not the overall data.

        Args:
            index (Union[list,int]): one or many indices to label.
            value (Optional[Any]): The label value. If not provided, no modification
                                    to the underlying dataset is done.
        """
        if isinstance(index, int):
            index = [index]
        if not isinstance(value, (list, tuple)):
            value = [value]
        indexes = self._pool_to_oracle_index(index)
        active_step = self.current_al_step + 1
        for index, val in zip_longest(indexes, value, fillvalue=None):
            self.labelled_map[index] = active_step



import warnings
from typing import Union, List, Optional, Any, TYPE_CHECKING, Protocol, Tuple

import numpy as np
from sklearn.utils import check_random_state
from torch.utils import data as torchdata

from baal.utils.equality import assert_not_none


class SizeableDataset(torchdata.Dataset):
    def __len__(self):
        pass


if TYPE_CHECKING:
    Dataset = SizeableDataset
else:
    Dataset = torchdata.Dataset


class SplittedDataset(Dataset):
    """Abstract class for Dataset that can be splitted.

    Args:
        labelled: An array that acts as a mask which is greater than 1 for every
            data point that is labelled, and 0 for every data point that is not
            labelled.
        random_state: Set the random seed for label_randomly().
        last_active_steps: If specified, will iterate over the last active steps
                            instead of the full dataset. Useful when doing partial finetuning.
    """

    def __init__(
        self,
        labelled,
        random_state=None,
        last_active_steps: int = -1,
    ) -> None:
        self.labelled_map = labelled
        self.random_state = check_random_state(random_state)
        if last_active_steps == 0 or last_active_steps < -1:
            raise ValueError("last_active_steps must be > 0 or -1 when disabled.")
        self.last_active_steps = last_active_steps
        self._indices_cache: Tuple[int, List[int]] = (-1, [])

    def get_indices_for_active_step(self) -> List[int]:
        """Returns the indices required for the active step.

        Returns the indices of the labelled items. Also takes into account self.last_active_step.

        Returns:
            List of the selected indices for training.
        """
        if (curr_al_step := self.current_al_step) != self._indices_cache[0]:
            if self.last_active_steps == -1:
                min_labelled_step = 0
            else:
                min_labelled_step = max(0, curr_al_step - self.last_active_steps)

            # we need to work with lists since arrow dataset is not compatible with np.int types!
            indices = [
                indx for indx, val in enumerate(self.labelled_map) if val > min_labelled_step
            ]
            self._indices_cache = (curr_al_step, indices)
        return self._indices_cache[1]

    def is_labelled(self, idx: int) -> bool:
        """Check if a datapoint is labelled."""
        return bool(self.labelled[idx].item() == 1)

    def __len__(self) -> int:
        """Return how many actual data / label pairs we have."""
        return len(self.get_indices_for_active_step())

    def __getitem__(self, index):
        raise NotImplementedError

    @property
    def n_unlabelled(self):
        """The number of unlabelled data points."""
        return (~self.labelled).sum()

    @property
    def n_labelled(self):
        """The number of labelled data points."""
        return self.labelled.sum()

    def label(self, index: Union[list, int], value: Optional[Any] = None) -> None:
        """
        Label data points.
        The index should be relative to the pool, not the overall data.

        Args:
            index: one or many indices to label.
            value: The label value. If not provided, no modification
                                    to the underlying dataset is done.
        """
        raise NotImplementedError

    def label_randomly(self, n: int = 1) -> None:
        """
        Label `n` data-points randomly.

        Args:
            n (int): Number of samples to label.
        """
        self.label(self.random_state.choice(self.n_unlabelled, n, replace=False).tolist())

    @property
    def _labelled(self):
        warnings.warn(
            "_labelled as been renamed labelled. Please update your script.", DeprecationWarning
        )
        return self.labelled

    @property
    def current_al_step(self) -> int:
        """Get the current active learning step."""
        return int(self.labelled_map.max())

    @property
    def labelled(self):
        """An array that acts as a boolean mask which is True for every
        data point that is labelled, and False for every data point that is not
        labelled."""
        return self.labelled_map.astype(bool)

    def _labelled_to_oracle_index(self, index: int) -> int:
        return int(self.labelled.nonzero()[0][index].squeeze().item())

    def _pool_to_oracle_index(self, index: Union[int, List[int]]) -> List[int]:
        if isinstance(index, np.int64) or isinstance(index, int):
            index = [index]

        lbl_nz = (~self.labelled).nonzero()[0]
        return [int(lbl_nz[idx].squeeze().item()) for idx in index]

    def _oracle_to_pool_index(self, index: Union[int, List[int]]) -> List[int]:
        if isinstance(index, int):
            index = [index]

        # Pool indices are the unlabelled, starts at 0
        lbl_cs = np.cumsum(~self.labelled) - 1
        return [int(lbl_cs[idx].squeeze().item()) for idx in index]
